#include <stdafx.h>
#include "httpclient.h"
#include <string>
#include "WriteBuffer.h"
#include "checkinexecution.h"

#define MAX_ALLOC_SIZE	(1024 * 10000)	// LocalAlloc() ִ밪. Ƿ 10MB Ҵ.

namespace httpbase
{

static LPSTR  Encode_utf8(LPCWSTR wStr)     // UNICODE -> UTF8
{
	int size = WideCharToMultiByte(CP_UTF8, 0, wStr, -1, NULL, 0, NULL, NULL);         
	LPSTR utf8Str = new char[size];    
	WideCharToMultiByte(CP_UTF8, 0, wStr, -1, utf8Str, size, 0, 0);    

	return utf8Str;
}

static LPWSTR Decode_utf8(LPCSTR utf8Str) // UTF8 -> UNICODE
{
	int size = MultiByteToWideChar(CP_UTF8, 0,  utf8Str, -1, NULL, 0);
	LPWSTR wStr = new WCHAR[size];
	MultiByteToWideChar(CP_UTF8, 0,  utf8Str, -1, wStr, size);

	return wStr;
}

const TCHAR *HTTP_MIME[]		= { _T("*/*"), _T("text/*"), NULL};
const TCHAR *HTTPS_STRING		= _T("https");

GenericHTTPClient::GenericHTTPClient()
	: m_dwHttpResponseCode ( 0 )
{
	m_http_open				= NULL;
	m_http_connection		= NULL;
	m_http_request			= NULL;
	m_http_response_header	= NULL;
	m_http_response_html	= NULL;
	m_content_length		= (DWORD)-1;
	m_callback = NULL;

	m_bIsNclicks = false;
}

GenericHTTPClient::~GenericHTTPClient()
{
	m_http_open			= NULL;
	m_http_connection	= NULL;
	m_http_request		= NULL;

	SAFE_TASKMEMDELETES(m_http_response_header);
	SAFE_TASKMEMDELETES(m_http_response_html);
}

DWORD GenericHTTPClient::GetContentLength()
{
	return m_content_length;
}

GenericHTTPClient::RequestMethod GenericHTTPClient::GetMethod(int method)
{
	if( method > 0 )
	{
		return static_cast<GenericHTTPClient::RequestMethod>(method);
	}

	return GenericHTTPClient::RequestGetMethod;
}

GenericHTTPClient::TypePostArgument GenericHTTPClient::GetPostArgumentType(int type)
{
	if( type > 0 )
	{
		return static_cast<GenericHTTPClient::TypePostArgument>(type);
	}

	return GenericHTTPClient::TypeNormal;
}

void GenericHTTPClient::DebugMessage()
{
	m_error = ::GetLastError();
#ifdef	_DEBUG
	LPVOID	msg_buffer;
	DWORD ret=FormatMessage( FORMAT_MESSAGE_ALLOCATE_BUFFER | FORMAT_MESSAGE_FROM_SYSTEM,
		NULL,
		::GetLastError(),
		MAKELANGID(LANG_NEUTRAL, SUBLANG_DEFAULT),
		reinterpret_cast<LPTSTR>(&msg_buffer),
		0,
		NULL);
	OutputDebugString(reinterpret_cast<LPTSTR>(msg_buffer));
//	MessageBox(NULL, (LPCWSTR)msg_buffer, _T("Error"), 0);
	LocalFree(msg_buffer);		
#endif
}

BOOL GenericHTTPClient::Connect( LPCTSTR cookies, LPCTSTR address, LPCTSTR agent,
								 WORD port, LPCTSTR id, LPCTSTR password, int timeout )
{
	m_http_open = ::InternetOpen( agent, INTERNET_OPEN_TYPE_PRECONFIG,
								_T(""), _T(""), 0);
	if( NULL == m_http_open )
	{
		DebugMessage();
		return FALSE;
	}

	m_http_connection = ::InternetConnect( m_http_open, address, port, id, password,
											INTERNET_SERVICE_HTTP,
											0,
											0);
	
	if( NULL == m_http_connection )
	{
		DebugMessage();
		::CloseHandle(m_http_open);
		return FALSE;
	}

	if( NULL != cookies && lstrcmp(cookies, _T("INVALID_COOKIE")) != 0 )
	{
		std::basic_string<TCHAR> url = _T("http://");
		url += address;
		BOOL ret = ::InternetSetCookie( url.c_str(), NULL, cookies );
	}

	return TRUE;
}

BOOL GenericHTTPClient::Close()
{
	if( NULL != m_http_request )
	{
		::InternetCloseHandle(m_http_request);
	}
	if( NULL != m_http_connection )
	{
		::InternetCloseHandle(m_http_connection);
	}
	if( NULL != m_http_open )
	{
		::InternetCloseHandle(m_http_open);
	}

	return TRUE;
}

void GenericHTTPClient::InitializePostArguments()
{
	m_arguments.clear();
}

void GenericHTTPClient::AddPostArguments( LPCTSTR name, DWORD value )
{
	GenericHTTPArgument arg;
	::FillMemory( &arg, sizeof(arg), 0x00 );

	lstrcpy( arg.name, name	);
	wsprintf( arg.value, _T("%d"), value );
	arg.type = GenericHTTPClient::TypeNormal;
	m_arguments.push_back(arg);	
}

void GenericHTTPClient::AddPostArguments( LPCTSTR name, LPCTSTR value, BOOL is_binary )
{
	GenericHTTPArgument arg;
	::ZeroMemory( &arg, sizeof(arg) );

	lstrcpy( arg.name, name );
	lstrcpy( arg.value, value );

	arg.type = GenericHTTPClient::TypeNormal;
	if( is_binary )
	{
		/// this case, value is filename.
		arg.type = GenericHTTPClient::TypeBinary;
	}

	m_arguments.push_back(arg);
}

BOOL GenericHTTPClient::Request(LPCTSTR cookie, LPCTSTR url, int method, LPCTSTR filename, LPCTSTR agent, IFileDownloadCallBack *callback )
{
	BOOL ret = TRUE;
	DWORD port = 0;
	TCHAR protocol[SIZE_HEADER_BUFFER] = _T("");
	TCHAR address[SIZE_HEADER_BUFFER] = _T("");
	TCHAR uri[SIZE_HEADER_BUFFER] = _T("");

	SAFE_TASKMEMDELETES(m_http_response_header);
	SAFE_TASKMEMDELETES(m_http_response_html);

	m_result_size = 0;
	m_callback = callback;

	ParseHttpURL( url, protocol, address, port, uri );
	if( Connect( cookie, address, agent, (WORD)port ) == FALSE )
		return FALSE;

	ret = RequestOfURI( uri, method );
	if( TRUE == ret )
	{
		ret = Response( filename, m_result_size, &m_http_response_html, &m_http_response_header );
	}

	// http ڵ带 ´.
	DWORD dwStatus = 0 ;
	DWORD cbStatus = sizeof(DWORD) ;
	if (::HttpQueryInfo(m_http_request, HTTP_QUERY_STATUS_CODE | HTTP_QUERY_FLAG_NUMBER, &dwStatus, &cbStatus, NULL))
	{
		m_dwHttpResponseCode = dwStatus;
	}

	Close();
	return ret;
}

void GenericHTTPClient::ParseHttpURL(LPCTSTR url, LPTSTR protocol, LPTSTR address, DWORD& port, LPTSTR uri)
{
	int position = 0;
	BOOL flag = FALSE;

	while( lstrlen(url) > 0 
			&& position < lstrlen(url)
			&&  wcsncmp((url+position),_T(":"), 1))
	{
		++position;
	}

	if (!wcsncmp( (url+position+1), _T("/"), 1))
	{
		if( NULL != protocol )
		{
			wcsncpy_s( protocol, SIZE_HEADER_BUFFER, url, position );
			protocol[position] = 0;
		}
		flag = TRUE;
	}
	else
	{
		if( NULL != protocol )
		{
			wcsncpy_s( protocol,SIZE_HEADER_BUFFER, _T("http"), 4);
			protocol[4] = 0;
		}
	}

	int start_position = 0;
	if( flag )
	{
		position += 3;
		start_position = position;
	}
	else
	{
		start_position = position = 0;
	}

	flag = FALSE;
	while( lstrlen(url) > 0 
		&& position < lstrlen(url)
		&&  wcsncmp((url+position),_T("/"), 1))
	{
		++position;
	}

	int find_position = start_position;

	for( ;find_position <= position; find_position++ )
	{
		if( !wcsncmp( (url + find_position), _T(":"), 1))
		{
			flag = TRUE;
			break;
		}
	}

	if( TRUE == flag )
	{
		TCHAR tmp[SIZE_BUFFER] = _T("");
		wcsncpy_s( tmp, SIZE_BUFFER, (url + find_position + 1), position - find_position );
		port = _tstol(tmp);
		wcsncpy_s( address, SIZE_HEADER_BUFFER, (url+start_position), find_position-start_position);
	}
	else if( !lstrcmp( protocol, HTTPS_STRING ))
	{
		port = INTERNET_DEFAULT_HTTPS_PORT;
		wcsncpy_s( address, SIZE_HEADER_BUFFER, (url+start_position), position-start_position);
	}
	else
	{
		port = INTERNET_DEFAULT_HTTP_PORT;
		wcsncpy_s( address, SIZE_HEADER_BUFFER, (url+start_position), position-start_position);
	}

	uri[0] = 0;
	if( position < lstrlen( url ) )
	{
		wcsncpy_s( uri, SIZE_HEADER_BUFFER, (url+position), lstrlen(url)-position);
	}
	return;
}

DWORD GenericHTTPClient::GetReponseSize()
{
	TCHAR content_size[32];
	DWORD content_size_length = sizeof(content_size);
	if( NULL == ::HttpQueryInfo( m_http_request, HTTP_QUERY_CONTENT_LENGTH , content_size,
		&content_size_length, NULL))
	{
		DebugMessage();
		return (DWORD)-1;
	}

	return (DWORD)_tstol(content_size);
}

BOOL GenericHTTPClient::SaveToBuffer( DWORD &recvSize, DWORD contentLength, IWriteBuffer* writeBuffer )
{
	DWORD readBytes;
	BYTE response_buffer[SIZE_BUFFER];

	int retSize;
	while((readBytes=ResponseOfBytes(response_buffer, SIZE_BUFFER, recvSize, contentLength)) != NULL)
	{
		retSize = writeBuffer->Write( response_buffer, readBytes	);
		if( -1 == retSize )
		{
			DebugMessage();
			return FALSE;
		}

		//Sleep(1000);

		recvSize+=readBytes;
		if( m_callback )
		{
			if( -1 == m_callback->Readed(recvSize, contentLength) )
				return FALSE;
		}
	}
	

	return TRUE;
}

BOOL GenericHTTPClient::GetHttpQueryHeaders(TCHAR **requestHeaders)
{
	int expectSize = 10240;
	DWORD readedHeaderSize;
	BOOL ret = FALSE;

	for( int i = 0; i < 3; i++ )
	{
		*requestHeaders = (TCHAR *)CoTaskMemAlloc(expectSize);
		if( NULL == *requestHeaders )
		{
			return FALSE;
		}

		readedHeaderSize = expectSize;
		ret = ::HttpQueryInfo( m_http_request, HTTP_QUERY_RAW_HEADERS_CRLF, 
								 *requestHeaders, &readedHeaderSize, NULL);

		if( TRUE == ret )
		{
			break;
		}

		SAFE_TASKMEMDELETES(*requestHeaders);

		if( ::GetLastError() != ERROR_INSUFFICIENT_BUFFER )
		{
			return FALSE;
		}

		expectSize *= 10;
	}
	return ret;
}

BOOL GenericHTTPClient::CheckReceiveSize( int filesize, int recvsize )
{
	if( filesize != (DWORD)-1 && filesize != recvsize )
	{
		return FALSE;
	}

	if( filesize == (DWORD)-1 && recvsize == 0 )
	{
		return FALSE;
	}

	return TRUE;
}

IWriteBuffer* GenericHTTPClient::MakeWriteBuffer( LPCTSTR filename, DWORD content_length, BYTE **requestBody )
{
	IWriteBuffer* writeBuffer;
	if( NULL != filename )
	{
		writeBuffer = new WriteFileBuffer(filename);
	}
	else
	{
		if( NULL == requestBody )
		{
			return NULL;
		}

		if( (DWORD)-1 == content_length )
		{
			writeBuffer = new WriteAutoRaisedMemoryBuffer(requestBody);
		}
		else
		{
			writeBuffer = new WritePreparedMemoryBuffer(content_length, requestBody);
		}
	}

	if( NULL != writeBuffer && FALSE == writeBuffer->Open() )
	{
		SAFE_DELETE(writeBuffer);
	}
	return writeBuffer;
}

BOOL GenericHTTPClient::Response(LPCTSTR filename, DWORD &result_size, BYTE** requestBody, TCHAR** requestHeader)
{
	if( NULL == m_http_request )
	{
		DebugMessage();
		return FALSE;
	}

	m_content_length = GetReponseSize();

	IWriteBuffer *writeBuffer = MakeWriteBuffer( filename, m_content_length, requestBody );
	if( NULL == writeBuffer || FALSE == SaveToBuffer( result_size, m_content_length, writeBuffer ) )
	{
		DebugMessage();
		SAFE_DELETE(writeBuffer);

		return FALSE;
	}

	BOOL ret = TRUE;
	ret = writeBuffer->Close();
	delete writeBuffer;
	if( FALSE == ret )
	{
		return FALSE;
	}

	if( FALSE == GetHttpQueryHeaders(requestHeader) )
	{
		DebugMessage();
		return FALSE;
	}

	return CheckReceiveSize( m_content_length, result_size );
}

DWORD GenericHTTPClient::ResponseOfBytes(PBYTE buffer, DWORD size, DWORD currentSize, DWORD filesize)

{
	DWORD bytes = 0;
	if( NULL == m_http_request )
	{
		DebugMessage();
		return 0;
	}

	DWORD dwRead;
	BOOL bOK = InternetReadFile(m_http_request, buffer, size, &dwRead ); 
	if (FALSE == bOK)
	{
		DebugMessage();
		return 0;
	}

	return dwRead;
}

BYTE* GenericHTTPClient::QueryHTTPResponse()
{
	return m_http_response_html;
}

DWORD	GenericHTTPClient::QueryHTTPResponseSize()
{
	return m_result_size;
}

LPCTSTR GenericHTTPClient::QueryHTTPResponseHeader()
{
	return m_http_response_header;
}

DWORD GenericHTTPClient::GetLastError()
{
	return m_error;
}

void GenericHTTPClient::GetContentType(LPCTSTR name, LPTSTR type)
{
	LONG	len=1024;
	DWORD	dot=0;

	for(dot=lstrlen(name);dot>0;dot--)
	{
		if(!_tcsncmp((name+dot), _T("."), 1))
			break;
	}

	HKEY	hKEY;
	LPTSTR	word=(TCHAR*)(name+dot);

	if(RegOpenKeyEx(HKEY_CLASSES_ROOT, word, 0, KEY_QUERY_VALUE, &hKEY)==ERROR_SUCCESS)
	{
		if(RegQueryValueEx(hKEY, _T("Content Type"), NULL, NULL, (LPBYTE)type, (unsigned long*)&len)!=ERROR_SUCCESS)
			wcsncpy_s(type, 1024, _T("application/octet-stream"), lstrlen(_T("application/octet-stream")));
		RegCloseKey(hKEY);
	}
	else
	{
		wcsncpy_s(type, 1024,  _T("application/octet-stream"), lstrlen(_T("application/octet-stream")));
	}
}

BOOL GenericHTTPClient::RequestOfURI( LPCTSTR uri, int method )
{
	BOOL ret=TRUE;

	try
	{
		switch(method)
		{
		case	GenericHTTPClient::RequestGetMethod:
		default:
			ret=RequestGet(uri);
			break;
		case	GenericHTTPClient::RequestPostMethod:		
			ret=RequestPost(uri);
			break;
		case	GenericHTTPClient::RequestPostMethodMultiPartsFormData:
			ret=RequestPostMultiPartsFormData(uri);
			break;
		}
	}
	catch(...)
	{
	}

	return ret;
}

BOOL GenericHTTPClient::RequestGet(LPCTSTR uri)
{
	m_http_request = ::HttpOpenRequest( m_http_connection, HTTP_VERB_GET, uri, HTTP_VERSION,
										NULL, &HTTP_MIME[1], 
										INTERNET_FLAG_PRAGMA_NOCACHE|INTERNET_FLAG_NO_CACHE_WRITE|INTERNET_FLAG_RELOAD,
										0 );
	if( NULL == m_http_request )
	{
		DebugMessage();
		return FALSE;
	}
	if( m_bIsNclicks )
	{
		BOOL bResult = ::HttpAddRequestHeaders( m_http_request, HTTP_CC_HOST, lstrlen(HTTP_CC_HOST),
			HTTP_ADDREQ_FLAG_REPLACE | HTTP_ADDREQ_FLAG_ADD );
		bResult = ::HttpAddRequestHeaders( m_http_request, HTTP_REFER_NAME, lstrlen(HTTP_REFER_NAME),
			HTTP_ADDREQ_FLAG_REPLACE | HTTP_ADDREQ_FLAG_ADD );
	}
	if ( NULL == ::HttpAddRequestHeaders( m_http_request, HTTP_ACCEPT, lstrlen(HTTP_ACCEPT),
											HTTP_ADDREQ_FLAG_REPLACE))
	{
		DebugMessage();
		return FALSE;
	}

	if( NULL == ::HttpSendRequest( m_http_request, NULL, 0, NULL, 0 ))
	{
		DebugMessage();
		return FALSE;
	}

	return TRUE;
}

BOOL GenericHTTPClient::RequestPostMultiPartsFormData(LPCTSTR uri)
{
	LPCTSTR content_type = _T("Content-Type: multipart/form-data; boundary=--MULTI-PARTS-FORM-DATA-BOUNDARY\r\n");
	PBYTE post_buffer = NULL;
	DWORD post_buffer_length = AllocMultiPartsFormData(post_buffer, _T("--MULTI-PARTS-FORM-DATA-BOUNDARY"));

	if (post_buffer_length == 0)
	{
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

	m_http_request =::HttpOpenRequest(	m_http_connection,
										HTTP_VERB_POST, 
										uri, 
										HTTP_VERSION, 
										_T(""), 
										&HTTP_MIME[1],
										INTERNET_FLAG_KEEP_CONNECTION | INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_FORMS_SUBMIT,	
										0);

	if( NULL == m_http_request )
	{
		DebugMessage();
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

	if( NULL == HttpAddRequestHeaders( m_http_request, HTTP_ACCEPT, lstrlen(HTTP_ACCEPT), HTTP_ADDREQ_FLAG_REPLACE))
	{
		DebugMessage();
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

	if( NULL == ::HttpAddRequestHeaders( m_http_request, content_type, lstrlen(content_type), 
										HTTP_ADDREQ_FLAG_ADD_IF_NEW))
	{
		DebugMessage();
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

#ifdef	_DEBUG
	TCHAR	http_request[SIZE_HTTP_BUFFER]= _T("");
	DWORD	http_request_length=SIZE_HTTP_BUFFER;

	::ZeroMemory(http_request, SIZE_HTTP_BUFFER);
	if(!::HttpQueryInfo(m_http_request, HTTP_QUERY_RAW_HEADERS_CRLF, http_request, 
						&http_request_length, NULL))
	{
		DebugMessage();
		//return FALSE;
	}

	ATLTRACE(http_request);
#endif

	TCHAR	content_length[SIZE_BUFFER] = _T("");
	::ZeroMemory(content_length, SIZE_BUFFER);

	wsprintf( content_length, _T("Content-Length: %d\r\n"), post_buffer_length);

	if(!::HttpAddRequestHeaders( m_http_request, content_length, lstrlen(content_length), HTTP_ADDREQ_FLAG_ADD_IF_NEW))
	{
		DebugMessage();
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

	INTERNET_BUFFERS internet_buffer_in={0};
	internet_buffer_in.dwStructSize=sizeof(INTERNET_BUFFERS);
	internet_buffer_in.Next=NULL;	

	if(!::HttpSendRequestEx(m_http_request, &internet_buffer_in, NULL, HSR_INITIATE, 0))
	{
		DebugMessage();
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

	DWORD out_post_buffer_length=0;
	if(!::InternetWriteFile(m_http_request, post_buffer, post_buffer_length, &out_post_buffer_length))
	{
		DebugMessage();
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

	if(!::HttpEndRequest(m_http_request, NULL, HSR_INITIATE, 0))
	{
		DebugMessage();
		FreeMultiPartsFormData(post_buffer);
		return FALSE;
	}

	// FREE POST MULTI-PARTS FORM DATA ARGUMENTS
	FreeMultiPartsFormData(post_buffer);

	return TRUE;
}


DWORD GenericHTTPClient::GetPostArguments( BYTE* arguments, DWORD length )
{
	std::vector<GenericHTTPArgument>::iterator itArg;

	::ZeroMemory( arguments, length );

	for( itArg = m_arguments.begin(); itArg < m_arguments.end(); )
	{
		char *name =		Encode_utf8(itArg->name);
		char *seperate =	Encode_utf8(_T("="));
		char *value =		Encode_utf8(itArg->value);

		strcat_s( (char *)arguments, SIZE_BUFFER, name );
		strcat_s( (char *)arguments, SIZE_BUFFER, seperate );
		strcat_s( (char *)arguments, SIZE_BUFFER, value );

		delete [] name;
		delete [] seperate;
		delete [] value;

		if( (++itArg) < m_arguments.end() )
		{
			char *serperate2 = Encode_utf8(_T("&"));
			strcat_s( (char *)arguments, SIZE_BUFFER, serperate2 );
			delete [] serperate2;
		}
	}

	*(arguments+length-1) = '\0';
	return strlen((const char *)arguments);
}


BOOL GenericHTTPClient::RequestPost(LPCTSTR uri)
{
	BYTE post_arguments[SIZE_BUFFER]= "";
	LPCTSTR content_type= _T("Content-Type: application/x-www-form-urlencoded\r\n");

	DWORD post_arguments_length = GetPostArguments(post_arguments, SIZE_BUFFER);
	m_http_request =::HttpOpenRequest( m_http_connection, HTTP_VERB_POST, uri, HTTP_VERSION, 
										_T(""), &HTTP_MIME[1], 
										INTERNET_FLAG_KEEP_CONNECTION | INTERNET_FLAG_NO_CACHE_WRITE | INTERNET_FLAG_FORMS_SUBMIT,
										0);

	if( NULL == m_http_request )
	{
		DebugMessage();
		return FALSE;
	}

	if ( NULL == ::HttpAddRequestHeaders( m_http_request, HTTP_ACCEPT, lstrlen(HTTP_ACCEPT),
		HTTP_ADDREQ_FLAG_REPLACE))
	{
		DebugMessage();
		return FALSE;
	}

	if( NULL == ::HttpSendRequest( m_http_request, content_type, lstrlen(content_type),
									(LPVOID)post_arguments, strlen((const char *)post_arguments)))
	{
		DebugMessage();
		return FALSE;
	}

	return TRUE;
}

DWORD GenericHTTPClient::AllocMultiPartsFormData(PBYTE& in_buffer, LPCTSTR boundary)
{
	if (NULL != in_buffer)
	{
		::LocalFree(in_buffer);
		in_buffer = NULL;
	}

	SIZE_T size = GetMultiPartsFormDataLength();

	if (size > MAX_ALLOC_SIZE)
		return 0;

	in_buffer = (PBYTE)::LocalAlloc(LPTR, size);

	// Handle error condition
	if (NULL == in_buffer)
	{
		CString strError;
		strError.Format(_T("LocalAlloc failed (%d)\n"), GetLastError());
		AfxMessageBox(strError);

		return 0;
	}

	std::vector<GenericHTTPArgument>::iterator itArgv;

	DWORD position=0;
	DWORD buffer_size=0;
	TCHAR tmp_buffer[10240];
	char *tmp_buffer2= NULL;

	for (itArgv = m_arguments.begin(); itArgv < m_arguments.end(); ++itArgv)
	{
		BYTE* buffer= NULL;
		tmp_buffer[0] = '\0';
		switch( itArgv->type )
		{
			case GenericHTTPClient::TypeNormal:
				buffer = (BYTE *)::LocalAlloc(LPTR, SIZE_HTTP_HEAD_LINE * 4);

				if (NULL == buffer)
				{
					CString strError;
					strError.Format(_T("LocalAlloc failed (%d)\n"), GetLastError());
					AfxMessageBox(strError);

					return 0;
				}

				wsprintf(tmp_buffer, 
					_T("--%s\r\n")
					_T("Content-Disposition: form-data; name=\"%s\"\r\n")
					_T("\r\n")
					_T("%s\r\n"),
					boundary,
					itArgv->name,
					itArgv->value);

				tmp_buffer2 = (char *)Encode_utf8(tmp_buffer);
				strcpy_s( (char *)buffer, SIZE_HTTP_HEAD_LINE*4, tmp_buffer2 );
				delete [] tmp_buffer2;
				tmp_buffer2 = NULL;
				buffer_size = strlen( (const char *)buffer );
				break;

			case GenericHTTPClient::TypeBinary:
				DWORD read_bytes = 0;
				DWORD total_bytes = 0;
			
				HANDLE hFile=::CreateFile(itArgv->value, 
											GENERIC_READ, 
											FILE_SHARE_READ, 
											NULL, 
											OPEN_EXISTING, 
											FILE_ATTRIBUTE_NORMAL | FILE_FLAG_SEQUENTIAL_SCAN, 
											NULL); 

				DWORD	size = ::GetFileSize(hFile, NULL);
				SIZE_T	allocSize = (SIZE_HTTP_HEAD_LINE*3+size+1) * sizeof(TCHAR);

				if (allocSize > MAX_ALLOC_SIZE)
				{
					::CloseHandle(hFile);
					return 0;
				}

				buffer = (BYTE *)::LocalAlloc(LPTR, allocSize);

				if (NULL == buffer)
				{
					CString strError;
					strError.Format(_T("LocalAlloc failed (%d)\n"), GetLastError());
					AfxMessageBox(strError);

					return 0;
				}


				BYTE	bytes[SIZE_BUFFER+1]= "";
				TCHAR	type[1024];
				GetContentType(itArgv->value, type);

				wsprintf(tmp_buffer,
						_T("--%s\r\n")
						_T("Content-Disposition: form-data; name=\"%s\"; filename=\"%s\"\r\n")
						_T("Content-Type: %s\r\n")
						_T("\r\n"),
						boundary,
						itArgv->name, itArgv->value,
						type
					);

				tmp_buffer2 = (char *)Encode_utf8(tmp_buffer);
				strcpy_s( (char *)buffer, (SIZE_HTTP_HEAD_LINE*3+size+1) * sizeof(TCHAR), tmp_buffer2 );
				delete [] tmp_buffer2;
				tmp_buffer2 = NULL;

				DWORD	buffer_position= strlen((const char *)buffer);

				while(::ReadFile(hFile, bytes, SIZE_BUFFER, &read_bytes, NULL) && read_bytes>0 && total_bytes<=size){
					::CopyMemory((buffer+buffer_position+total_bytes), bytes, read_bytes);
					::ZeroMemory(bytes, SIZE_BUFFER+1);
					total_bytes+=read_bytes;				
				}

				::CopyMemory((buffer+buffer_position+total_bytes), "\r\n", strlen("\r\n") * sizeof(char));
				::CloseHandle(hFile);

				buffer_size = buffer_position + total_bytes + strlen("\r\n") * sizeof(char);
				break;		
		}

		if (buffer && buffer_size > 0)
		{
			::CopyMemory((in_buffer + position), buffer, buffer_size);
			position += buffer_size;

			::LocalFree(buffer);
			buffer = NULL;
		}
	}

	char *utf8_boundary = Encode_utf8(boundary);
	int utf8_boudndary_size = strlen(utf8_boundary);
	::CopyMemory( (in_buffer+position), "--", 2 * sizeof(char));
	::CopyMemory( (in_buffer+position + (2 * sizeof(char)) ), utf8_boundary, utf8_boudndary_size * sizeof(char));
	::CopyMemory( (in_buffer+position + ( 2 +utf8_boudndary_size ) * sizeof(char) ), _T("--\r\n"), 3 * sizeof(char));

	//  PLANHOOD ߴ ڵ
	//::CopyMemory((pInBuffer+position+2+strlen(szBoundary)), "--\r\n", 4);
	delete [] utf8_boundary;
	return position + ( ( 2 + utf8_boudndary_size + 3 ) * sizeof(char) );
}

void GenericHTTPClient::FreeMultiPartsFormData(PBYTE& buffer)
{
	if (buffer)
	{
		::LocalFree(buffer);
		buffer = NULL;
	}

	return;
}

DWORD GenericHTTPClient::GetMultiPartsFormDataLength()
{
	std::vector<GenericHTTPArgument>::iterator itArgv;

	DWORD dwLen = 0;

	for (itArgv = m_arguments.begin(); itArgv < m_arguments.end(); ++itArgv)
	{
		switch (itArgv->type)
		{
		case GenericHTTPClient::TypeNormal:
			dwLen += SIZE_HTTP_HEAD_LINE * 4;
			break;

		case GenericHTTPClient::TypeBinary:
			HANDLE hFile = ::CreateFile(itArgv->value, 
				GENERIC_READ, // desired access
				FILE_SHARE_READ, // share mode
				NULL, // security attribute
				OPEN_EXISTING, // create disposition
				FILE_ATTRIBUTE_NORMAL, // flag and attributes
				NULL); // template file handle

			dwLen += SIZE_HTTP_HEAD_LINE * 3 + ::GetFileSize(hFile, NULL);
			::CloseHandle(hFile);
			break;
		}
	}

	return dwLen;
}

};